import uuid

from datetime import date, datetime
from typing import TYPE_CHECKING, Optional

from sqlalchemy import Boolean, Date, DateTime, Enum, ForeignKey, Unicode, Uuid
from sqlalchemy.orm import Mapped, mapped_column, relationship

from ..db.base_class import reg
from .voucher_type import VoucherType


if TYPE_CHECKING:
    from .employee_benefit import EmployeeBenefit
    from .incentive import Incentive
    from .inventory import Inventory
    from .journal import Journal
    from .user import User


@reg.mapped_as_dataclass(unsafe_hash=True)
class Voucher:
    __tablename__ = "vouchers"

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    date_: Mapped[date] = mapped_column("date", Date, nullable=False, index=True)
    narration: Mapped[str] = mapped_column(Unicode, nullable=False)
    is_reconciled: Mapped[bool] = mapped_column(Boolean, nullable=False)
    reconcile_date: Mapped[date] = mapped_column(Date, nullable=False)
    is_starred: Mapped[bool] = mapped_column(Boolean, nullable=False)
    creation_date: Mapped[datetime] = mapped_column(DateTime(), nullable=False)
    last_edit_date: Mapped[datetime] = mapped_column(DateTime(), nullable=False)
    voucher_type: Mapped[VoucherType] = mapped_column(Enum(VoucherType), nullable=False)
    user_id: Mapped[uuid.UUID] = mapped_column(Uuid, ForeignKey("users.id"), nullable=False)
    posted: Mapped[bool] = mapped_column("is_posted", Boolean, nullable=False)
    poster_id: Mapped[uuid.UUID | None] = mapped_column(Uuid, ForeignKey("users.id"), nullable=True)

    user: Mapped["User"] = relationship("User", foreign_keys=[user_id], back_populates="vouchers")
    poster: Mapped[Optional["User"]] = relationship("User", foreign_keys=[poster_id])

    journals: Mapped[list["Journal"]] = relationship(
        "Journal", cascade="delete, delete-orphan", back_populates="voucher"
    )

    inventories: Mapped[list["Inventory"]] = relationship(
        "Inventory", cascade="delete, delete-orphan", back_populates="voucher"
    )
    employee_benefits: Mapped[list["EmployeeBenefit"]] = relationship(
        "EmployeeBenefit", cascade="delete, delete-orphan", backref="voucher"
    )
    incentives: Mapped[list["Incentive"]] = relationship(
        "Incentive", cascade="delete, delete-orphan", backref="voucher"
    )

    def __init__(
        self,
        date_: date,
        voucher_type: VoucherType,
        user_id: uuid.UUID,
        narration: str = "",
        is_starred: bool = False,
        posted: bool = False,
        is_reconciled: bool = False,
        reconcile_date: date | None = None,
        poster_id: uuid.UUID | None = None,
        creation_date: datetime | None = None,
        last_edit_date: datetime | None = None,
    ):
        self.date_ = date_
        self.is_reconciled = is_reconciled
        self.reconcile_date = reconcile_date or date_
        self.is_starred = is_starred if is_starred is not None else False
        self.narration = narration
        self.posted = posted
        self.creation_date = creation_date or datetime.utcnow()
        self.last_edit_date = last_edit_date or datetime.utcnow()
        self.voucher_type = voucher_type
        self.user_id = user_id
        if poster_id is not None:
            self.poster_id = poster_id
