from datetime import date, timedelta
from typing import Sequence

from sqlalchemy import or_, select
from sqlalchemy.orm import Session

from ..models.db_setting import DbSetting
from ..models.setting_type import SettingType
from ..models.voucher_type import VoucherType
from ..schemas.settings import (
    AccountTypesSelected,
    LockDate,
    LockInformation,
    VoucherTypesSelected,
)


def get_lock_info(
    voucher_dates: Sequence[date], voucher_type: VoucherType, account_types: Sequence[int], db: Session
) -> tuple[bool, str]:
    date_ = date.today()
    start: date | None
    finish: date | None
    data: list[dict] = list(
        db.execute(
            select(DbSetting.data).where(
                DbSetting.setting_type == SettingType.VOUCHER_LOCK,
                or_(
                    DbSetting.valid_from == None,  # noqa: E711
                    DbSetting.valid_from <= date_,
                ),
                or_(
                    DbSetting.valid_till == None,  # noqa: E711
                    DbSetting.valid_till >= date_,
                ),
            )
        )
        .scalars()
        .all()
    )
    data = sorted(data, key=lambda d: d["index"], reverse=True)  # type: ignore[index]
    for it in data:
        li: LockInformation = LockInformation(
            voucher_types=[VoucherTypesSelected(id_=vt["id_"], name=vt["name"]) for vt in it["voucher_types"]],  # type: ignore[index]
            account_types=[AccountTypesSelected(id_=at["id_"], name=at["name"]) for at in it["account_types"]],  # type: ignore[index]
            start=LockDate(days=it["start"]["days"], date_=it["start"]["date_"]),  # type: ignore[index]
            finish=LockDate(days=it["finish"]["days"], date_=it["finish"]["date_"]),  # type: ignore[index]
            index=it["index"],  # type: ignore[index]
        )
        # Data format
        if len(li.voucher_types) != 0 and voucher_type not in [vt.id_ for vt in li.voucher_types]:
            continue
        if len(li.account_types) != 0 and len(set(account_types) & set([at.id_ for at in li.account_types])) == 0:
            continue

        if li.start.days is not None:
            start = date.today() - timedelta(days=li.start.days)
            if sum(1 for v in voucher_dates if v < start):
                return False, f"Vouchers before {start.strftime('%d-%b-%Y')} have been locked."
        if li.start.date_ is not None:
            start = li.start.date_
            if sum(1 for v in voucher_dates if v < start):
                return False, f"Vouchers before {start.strftime('%d-%b-%Y')} have been locked."

        if li.finish.days is not None:
            finish = date.today() + timedelta(days=li.finish.days)
            if sum(1 for v in voucher_dates if v > finish):
                return False, f"Vouchers after {finish.strftime('%d-%b-%Y')} have been locked."
        if li.finish.date_ is not None:
            finish = li.finish.date_
            if sum(1 for v in voucher_dates if v > finish):
                return False, f"Vouchers after {finish.strftime('%d-%b-%Y')} have been locked."

        return True, "Voucher allowed"
    return False, "Default Fallthrough"
