from datetime import datetime
from decimal import Decimal

import brewman.schemas.purchases as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session
from sqlalchemy.sql.expression import desc, func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.batch import Batch
from ...models.cost_centre import CostCentre
from ...models.inventory import Inventory
from ...models.journal import Journal
from ...models.product import Product
from ...models.stock_keeping_unit import StockKeepingUnit
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.Purchases)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchases"]),
) -> schemas.Purchases:
    return schemas.Purchases(
        start_date=get_start_date(request.session),
        finish_date=get_finish_date(request.session),
        body=[],
        footer=schemas.PurchasesItem(name="Total", quantity=Decimal(0), rate=Decimal(0), url=[], amount=Decimal(0)),
    )


@router.get("/{start}/{finish}", response_model=schemas.Purchases)
def report_data(
    start: str,
    finish: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["purchases"]),
) -> schemas.Purchases:
    with SessionFuture() as db:
        body, footer = build_report(start, finish, db)
        set_period(start, finish, request.session)
        return schemas.Purchases(
            start_date=start,
            finish_date=finish,
            body=body,
            footer=footer,
        )


def build_report(
    start_date: str, finish_date: str, db: Session
) -> tuple[list[schemas.PurchasesItem], schemas.PurchasesItem]:
    body = []
    quantity_sum = func.sum(Journal.debit * Inventory.quantity).label("quantity")
    amount_sum = func.sum(Journal.debit * Inventory.quantity * Inventory.rate * (1 + Inventory.tax)).label("amount")
    query = db.execute(
        select(Product, quantity_sum, amount_sum)
        .join(Voucher.journals)
        .join(Voucher.inventories)
        .join(Inventory.batch)
        .join(Batch.sku)
        .join(StockKeepingUnit.product)
        .where(
            Voucher.date_ >= datetime.strptime(start_date, "%d-%b-%Y"),
            Voucher.date_ <= datetime.strptime(finish_date, "%d-%b-%Y"),
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
            Journal.cost_centre_id == CostCentre.cost_centre_purchase(),
        )
        .group_by(Product)  # type: ignore
        .order_by(desc(amount_sum))
    ).all()
    total_amount = Decimal(0)
    for product, quantity, amount in query:
        rate = amount / quantity if quantity != 0 else 0
        total_amount += amount
        row = schemas.PurchasesItem(
            name=product.name,
            quantity=quantity,
            rate=rate,
            amount=amount,
            url=["/", "product-ledger", str(product.id)],
        )
        body.append(row)
    return (
        body,
        schemas.PurchasesItem(name="Total", quantity=Decimal(0), rate=Decimal(0), url=[], amount=total_amount),
    )
