from datetime import date, datetime

import brewman.schemas.trial_balance as schemas

from fastapi import APIRouter, Request, Security
from sqlalchemy import not_
from sqlalchemy.orm import Session, contains_eager, joinedload
from sqlalchemy.sql.expression import func, select

from ...core.security import get_current_active_user as get_user
from ...core.session import get_finish_date, get_start_date, set_period
from ...db.session import SessionFuture
from ...models.account_base import AccountBase
from ...models.account_type import AccountType
from ...models.journal import Journal
from ...models.voucher import Voucher
from ...models.voucher_type import VoucherType
from ...schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=schemas.TrialBalance)
def report_blank(
    request: Request,
    user: UserToken = Security(get_user, scopes=["trial-balance"]),
) -> schemas.TrialBalance:
    return schemas.TrialBalance(date_=get_finish_date(request.session), body=[])


@router.get("/{date_}", response_model=schemas.TrialBalance)
def report_data(
    date_: str,
    request: Request,
    user: UserToken = Security(get_user, scopes=["trial-balance"]),
) -> schemas.TrialBalance:
    set_period(get_start_date(request.session), date_, request.session)
    with SessionFuture() as db:
        return schemas.TrialBalance(
            date_=date_,
            body=build_report(datetime.strptime(date_, "%d-%b-%Y"), db),
        )


def build_report(date_: date, db: Session) -> list[schemas.TrialBalanceItem]:
    amount_sum = func.sum(Journal.amount * Journal.debit).label("amount")
    query = db.execute(
        select(AccountBase, amount_sum)
        .join(Journal.voucher)
        .join(Journal.account)
        .join(AccountBase.type_)
        .where(
            Voucher.date_ <= date_,
            not_(Voucher.voucher_type.in_([VoucherType.ISSUE, VoucherType.CLOSING_STOCK])),
        )
        .group_by(AccountBase, AccountType)  # type: ignore
        .order_by(AccountBase.type_id, func.abs(amount_sum).desc())
        .options(
            joinedload(AccountBase.type_, innerjoin=True),
            contains_eager(AccountBase.type_),
        )
    ).all()

    body = []
    for account, amount in query:
        if amount > 0:
            body.append(schemas.TrialBalanceItem(type_=account.type_.name, name=account.name, debit=amount))
        if amount < 0:
            body.append(schemas.TrialBalanceItem(type_=account.type_.name, name=account.name, credit=amount))
    return body
