import json
import uuid

from datetime import date, datetime
from decimal import Decimal

from fastapi import Form
from pydantic import BaseModel, ConfigDict, Field, field_validator, model_validator

from . import to_camel
from .account import AccountLink
from .cost_centre import CostCentreLink
from .employee_benefit import EmployeeBenefit
from .incentive import Incentive
from .inventory import Inventory
from .journal import Journal
from .voucher import VoucherIn


class JournalIn(VoucherIn):
    journals: list[Journal]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class PurchaseIn(VoucherIn):
    vendor: AccountLink
    inventories: list[Inventory]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class IssueIn(VoucherIn):
    source: CostCentreLink
    destination: CostCentreLink
    inventories: list[Inventory]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @model_validator(mode="after")
    def source_destination_unique(self) -> "IssueIn":
        if self.source.id_ == self.destination.id_:
            raise ValueError("Source and destination cannot be the same")
        return self

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class EmployeeBenefitIn(VoucherIn):
    employee_benefits: list[EmployeeBenefit]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class IncentiveIn(VoucherIn):
    incentives: list[Incentive]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @classmethod
    def load_form(cls, data: str = Form(...)) -> BaseModel:
        json_data = json.loads(data)
        return cls.parse_obj(json_data)


class IncentiveEmployee(BaseModel):
    id_: uuid.UUID | None = None
    employee_id: uuid.UUID
    cost_centre_id: uuid.UUID
    journal: Journal | None = None
    days_worked: Decimal = Field(ge=0)
    points: Decimal = Field(ge=0)
