import { AfterViewInit, Component, ElementRef, OnInit, OnDestroy, ViewChild } from '@angular/core';
import { FormControl, FormGroup } from '@angular/forms';
import { MatAutocompleteSelectedEvent } from '@angular/material/autocomplete';
import { MatDialog } from '@angular/material/dialog';
import { ActivatedRoute, Router } from '@angular/router';
import { Hotkey, HotkeysService } from 'angular2-hotkeys';
import { round } from 'mathjs';
import * as moment from 'moment';
import { BehaviorSubject, Observable, of as observableOf } from 'rxjs';
import { debounceTime, distinctUntilChanged, switchMap } from 'rxjs/operators';

import { AuthService } from '../auth/auth.service';
import { Account } from '../core/account';
import { AccountBalance } from '../core/account-balance';
import { AccountService } from '../core/account.service';
import { Batch } from '../core/batch';
import { DbFile } from '../core/db-file';
import { Inventory } from '../core/inventory';
import { Product } from '../core/product';
import { ProductSku } from '../core/product-sku';
import { ToasterService } from '../core/toaster.service';
import { User } from '../core/user';
import { Voucher } from '../core/voucher';
import { VoucherService } from '../core/voucher.service';
import { ProductService } from '../product/product.service';
import { ConfirmDialogComponent } from '../shared/confirm-dialog/confirm-dialog.component';
import { ImageDialogComponent } from '../shared/image-dialog/image-dialog.component';
import { ImageService } from '../shared/image.service';
import { MathService } from '../shared/math.service';

import { PurchaseDataSource } from './purchase-datasource';
import { PurchaseDialogComponent } from './purchase-dialog.component';

@Component({
  selector: 'app-purchase',
  templateUrl: './purchase.component.html',
  styleUrls: ['./purchase.component.css'],
})
export class PurchaseComponent implements OnInit, AfterViewInit, OnDestroy {
  @ViewChild('accountElement', { static: true }) accountElement?: ElementRef;
  @ViewChild('productElement', { static: true }) productElement?: ElementRef;
  @ViewChild('dateElement', { static: true }) dateElement?: ElementRef;
  public inventoryObservable = new BehaviorSubject<Inventory[]>([]);
  dataSource: PurchaseDataSource = new PurchaseDataSource(this.inventoryObservable);
  form: FormGroup<{
    date: FormControl<Date>;
    account: FormControl<string | null>;
    amount: FormControl<number>;
    addRow: FormGroup<{
      product: FormControl<string | null>;
      quantity: FormControl<string>;
      price: FormControl<string>;
      tax: FormControl<string>;
      discount: FormControl<string>;
    }>;
    narration: FormControl<string>;
  }>;

  voucher: Voucher = new Voucher();
  product: ProductSku | null = null;
  accBal: AccountBalance | null = null;

  displayedColumns = ['product', 'quantity', 'rate', 'tax', 'discount', 'amount', 'action'];

  accounts: Observable<Account[]>;
  products: Observable<ProductSku[]>;

  constructor(
    private route: ActivatedRoute,
    private router: Router,
    private dialog: MatDialog,
    private hotkeys: HotkeysService,
    private toaster: ToasterService,
    public auth: AuthService,
    private math: MathService,
    public image: ImageService,
    private ser: VoucherService,
    private productSer: ProductService,
    private accountSer: AccountService,
  ) {
    this.form = new FormGroup({
      date: new FormControl(new Date(), { nonNullable: true }),
      account: new FormControl<string | null>(null),
      amount: new FormControl({ value: 0, disabled: true }, { nonNullable: true }),
      addRow: new FormGroup({
        product: new FormControl(''),
        quantity: new FormControl('', { nonNullable: true }),
        price: new FormControl('', { nonNullable: true }),
        tax: new FormControl('', { nonNullable: true }),
        discount: new FormControl('', { nonNullable: true }),
      }),
      narration: new FormControl('', { nonNullable: true }),
    });
    this.accBal = null;
    // Listen to Account Autocomplete Change
    this.accounts = this.form.controls.account.valueChanges.pipe(
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) => (x === null ? observableOf([]) : this.accountSer.autocomplete(x))),
    );
    // Listen to Product Autocomplete Change
    this.products = this.form.controls.addRow.controls.product.valueChanges.pipe(
      debounceTime(150),
      distinctUntilChanged(),
      switchMap((x) =>
        x === null
          ? observableOf([])
          : this.productSer.autocompleteSku(
              x,
              true,
              moment(this.form.value.date).format('DD-MMM-YYYY'),
              // this.form.value.account?.id,
              '',
            ),
      ),
    );
  }

  ngOnInit() {
    this.route.data.subscribe((value) => {
      const data = value as { voucher: Voucher };

      this.loadVoucher(data.voucher);
    });
    this.hotkeys.add(
      new Hotkey(
        'f2',
        (): boolean => {
          setTimeout(() => {
            if (this.dateElement) {
              this.dateElement.nativeElement.focus();
            }
          }, 0);
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+s',
        (): boolean => {
          if (this.canSave()) {
            this.save();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
    this.hotkeys.add(
      new Hotkey(
        'ctrl+p',
        (): boolean => {
          if (this.voucher.id && !this.voucher.posted && this.auth.allowed('post-vouchers')) {
            this.post();
          }
          return false; // Prevent bubbling
        },
        ['INPUT', 'SELECT', 'TEXTAREA'],
      ),
    );
  }

  ngAfterViewInit() {
    this.focusAccount();
  }

  ngOnDestroy() {
    this.hotkeys.reset();
  }

  loadVoucher(voucher: Voucher) {
    this.voucher = voucher;
    this.form.setValue({
      date: moment(this.voucher.date, 'DD-MMM-YYYY').toDate(),
      account: this.voucher.vendor?.name ?? null,
      amount: Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)),
      addRow: {
        product: '',
        quantity: '',
        price: '',
        tax: '',
        discount: '',
      },
      narration: this.voucher.narration,
    });
    this.dataSource = new PurchaseDataSource(this.inventoryObservable);
    this.updateView();
  }

  focusAccount() {
    setTimeout(() => {
      if (this.accountElement) {
        this.accountElement.nativeElement.focus();
      }
    }, 0);
  }

  addRow() {
    const formValue = this.form.value.addRow;
    if (formValue === undefined) {
      return;
    }
    const quantity = this.math.parseAmount(formValue.quantity, 2);
    if (this.product === null || quantity <= 0) {
      return;
    }
    const price = this.product.isRateContracted
      ? (this.product.costPrice as number)
      : this.math.parseAmount(formValue.price, 2);
    const tax = this.product.isRateContracted ? 0 : this.math.parseAmount(formValue.tax, 5);
    const discount = this.product.isRateContracted ? 0 : this.math.parseAmount(formValue.discount, 5);
    if ((price as number) <= 0 || (tax as number) < 0 || (discount as number) < 0) {
      return;
    }
    const oldFiltered = this.voucher.inventories.filter((x) => x.batch?.sku.id === (this.product as ProductSku).id);
    if (oldFiltered.length) {
      this.toaster.show('Danger', 'Product already added');
      return;
    }
    this.voucher.inventories.push(
      new Inventory({
        quantity,
        rate: price,
        tax,
        discount,
        amount: round(quantity * (price as number) * (1 + tax) * (1 - discount), 2),
        batch: new Batch({ sku: this.product }),
      }),
    );
    this.resetAddRow();
    this.updateView();
  }

  resetAddRow() {
    this.form.controls.addRow.reset();
    this.product = null;
    this.form.controls.addRow.controls.price.enable();
    this.form.controls.addRow.controls.tax.enable();
    this.form.controls.addRow.controls.discount.enable();
    setTimeout(() => {
      if (this.productElement) {
        this.productElement.nativeElement.focus();
      }
    }, 0);
  }

  updateView() {
    this.inventoryObservable.next(this.voucher.inventories);
    this.form.controls.amount.setValue(
      round(Math.abs(this.voucher.inventories.map((x) => x.amount).reduce((p, c) => p + c, 0)), 2),
    );
  }

  editRow(row: Inventory) {
    const dialogRef = this.dialog.open(PurchaseDialogComponent, {
      width: '750px',
      data: { inventory: { ...row } },
    });

    dialogRef.afterClosed().subscribe((result: boolean | Inventory) => {
      if (!result) {
        return;
      }
      const j = result as Inventory;
      if (
        j.batch?.sku.id !== row.batch?.sku.id &&
        this.voucher.inventories.filter((x) => x.batch?.sku.id === j.batch?.sku.id).length
      ) {
        return;
      }
      Object.assign(row, j);
      this.updateView();
    });
  }

  deleteRow(row: Inventory) {
    this.voucher.inventories.splice(this.voucher.inventories.indexOf(row), 1);
    this.updateView();
  }

  canSave() {
    if (!this.voucher.id) {
      return true;
    }
    if (this.voucher.posted && this.auth.allowed('edit-posted-vouchers')) {
      return true;
    }
    return this.voucher.user.id === (this.auth.user as User).id || this.auth.allowed("edit-other-user's-vouchers");
  }

  post() {
    this.ser.post(this.voucher.id as string).subscribe(
      (result) => {
        this.loadVoucher(result);
        this.toaster.show('Success', 'Voucher Posted');
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  save() {
    const voucher: Voucher = this.getVoucher();
    this.ser.saveOrUpdate(voucher).subscribe(
      (result) => {
        this.toaster.show('Success', '');
        if (voucher.id === result.id) {
          this.loadVoucher(result);
        } else {
          this.router.navigate(['/purchase', result.id]);
        }
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  getVoucher(): Voucher {
    const formModel = this.form.value;
    this.voucher.date = moment(formModel.date).format('DD-MMM-YYYY');
    if (formModel.account !== null && typeof formModel.account !== 'string') {
      this.voucher.vendor = formModel.account;
    }
    this.voucher.narration = formModel.narration ?? '';
    return this.voucher;
  }

  delete() {
    this.ser.delete(this.voucher.id as string).subscribe(
      () => {
        this.toaster.show('Success', '');
        this.router.navigate(['/purchase'], { replaceUrl: true });
      },
      (error) => {
        this.toaster.show('Danger', error);
      },
    );
  }

  confirmDelete(): void {
    const dialogRef = this.dialog.open(ConfirmDialogComponent, {
      width: '250px',
      data: { title: 'Delete Voucher?', content: 'Are you sure? This cannot be undone.' },
    });

    dialogRef.afterClosed().subscribe((result: boolean) => {
      if (result) {
        this.delete();
      }
    });
  }

  displayFn(item?: Account | Product | string): string {
    return !item ? '' : typeof item === 'string' ? item : item.name;
  }

  accountSelected(event: MatAutocompleteSelectedEvent): void {
    this.form.controls.account.setValue(event.option.value);
  }

  productSelected(event: MatAutocompleteSelectedEvent): void {
    const product: ProductSku = event.option.value;
    const addRowForm = this.form.controls.addRow;
    this.product = product;
    addRowForm.controls.price.setValue(`${product.costPrice}`);
    if (product.isRateContracted) {
      addRowForm.controls.price.disable();
      addRowForm.controls.tax.disable();
      addRowForm.controls.discount.disable();
      addRowForm.controls.tax.setValue('RC');
      addRowForm.controls.discount.setValue('RC');
    } else {
      addRowForm.controls.price.enable();
      addRowForm.controls.tax.enable();
      addRowForm.controls.discount.enable();
      addRowForm.controls.tax.setValue('');
      addRowForm.controls.discount.setValue('');
    }
  }

  zoomImage(file: DbFile) {
    this.dialog.open(ImageDialogComponent, {
      width: '750px',
      data: file.resized,
    });
  }

  deleteImage(file: DbFile) {
    const index = this.voucher.files.indexOf(file);
    this.voucher.files.splice(index, 1);
  }
}
